package uk.ac.ic.doc.y2002.tool.editor;

import javax.swing.JLabel;
import java.awt.geom.*;
import java.awt.event.*;
import java.awt.image.*; 
import java.awt.*;
import java.util.*;
import org.jdom.*;
import uk.ac.ic.doc.y2002.tool.application.*;

/** The ArcLabel is the JLabel representing an arc on screen. 
 * Each arc is represented in the PNML. 
 */
public class ArcLabel extends PNLabel implements MouseListener, MouseMotionListener{
    /** Size of the small box in the centre of the arc which is the active (clickable) area. */
    int centresize = 20;
    /** NodeLabels referencing the source and target nodes of this arc. Set in the DisplayPane, when all the labels are initially loaded. */
    NodeLabel sourcelabel, targetlabel;
    /** Each ArcLabel has a LineLabel, which is a relatively simple subclass of JLabel that deals with drawing the lines. */
    LineLabel line;
    
    public ArcLabel(Element e) {
	super(e);
	// set up member variables
	offset = centresize/2;
	xoffset = yoffset = offset;
	preferredSize = new Dimension(centresize+textspace,centresize+textspace);
	// do custom initialisation
	int x = getJX();
	int y = getJY();
	setBounds(x-offset,y-offset,centresize+textspace,centresize+textspace);
	contextmenu = ArcContextMenu.class;
	propertiesbox = ArcPropertiesBox.class;
    }
    
    public void updateBounds() {
	setBounds(getX(),getY(),centresize+textspace,centresize+textspace);
    }
    
    public void paintComponent(Graphics g){
	super.paintComponent(g);
	Graphics2D g2 = (Graphics2D)g;
	g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
	
	/*// Draws the centre box of the arc - useful for debugging
	        g2.setStroke(new BasicStroke(1.0f));
		            g2.draw(new Rectangle2D.Double(0, 0, centresize, centresize));
			                g2.setPaint(fillcolour);
					      g2.fill (new Rectangle2D.Double(0, 0, centresize, centresize));
					                // done drawing centre box
							*/
	g2.setPaint(Color.black);
	
	// Draws the arc weight (if it is not 1 - default)
	String weight = String.valueOf(getWeight());
	if (getWeight()!=1)
	    g2.drawString(weight,centresize/2,centresize+10);
	
	// repaint lines
	line.updateLine();
    }
    
    /** When first drawn, the ArcLabel initialises its LineLabel and adds it to the DisplayPane.
     * It assumes that the sourcelabel and targetlabel attributes have been set.
     */
    public void dofirst() {
	super.dofirst();
	line = new LineLabel(sourcelabel,this,targetlabel);
	getParent().add(line);
	// draw the lines to source and target
	positionLines();
    }
    
    /** Calculates the position of the Arc - either snap-to-position or user specified.
     * Snap-to-position is when the arc automatically moves to the mid-point 
     * of its source and target.
     * User-specified is when the user chooses to move the arc elsewhere -
     * for example to move it out of the way of a complex part of the net.
     */
    public void positionLines(){
	if (getSnap()==true){
	    //System.out.println("ArcLabel:: snapping to position");
	    snapToPosition();
	} else {
	    //System.out.println("ArcLabel:: not snapping to position");
	    line.updateLine();
	}
    }
    
    /** Positions the arc exactly between its source and target. */
    protected void snapToPosition(){
	//System.out.println("ArcLabel::Snapping to position");
	int sx = sourcelabel.getJX();
	int sy = sourcelabel.getJY();
	int tx = targetlabel.getJX();
	int ty = targetlabel.getJY();
	
	int x = (sx + tx)/2;
	int y = (sy + ty)/2;
	
	setPosition(x-offset,y-offset);
	updateBounds();
	line.updateLine();
    }
    
    /** Handles the event of a user dragging the mouse while holding
     * down the mouse button.
     * Overridden to turn off auto-snap.
     */
    public void mouseDragged(MouseEvent e){
	super.mouseDragged(e);
	setSnap(false);
    }
    
    /** Updates the coordinates representing the location of the current label
     * in response to Mouse being dragged.
     * Overridden to deal with updating the LineLabel as well.
     */
    protected void updateLocation(MouseEvent e){
	super.updateLocation(e);
	line.updateLine();
	line.repaint();
    }
        
    /** Gets the Snap-to-position property in the JDOM Document.
     * This property is toolspecific.
     */
    public boolean getSnap(){
	Element toolspecific = getToolspecific();
	if (toolspecific!=null && toolspecific.getChildren("snap",ns).size()>0) {
	    String snap = toolspecific.getChildText("snap",ns);
	    if (snap.equals("true")) return true;
	    else return false;
	}
	return true; // default if element not found in the Document
    }
    
    /** Sets the Snap-to-position property in the JDOM Document.
     * This property is toolspecific.
     */
    public void setSnap(boolean snap){
	Element toolspecific = getToolspecific();
	if (toolspecific==null)
	    toolspecific = makeToolspecific();
	if (toolspecific.getChildren("snap",ns).size()<=0) {  
	    toolspecific.addContent(new Element("snap",ns));
	}
	toolspecific.getChild("snap",ns).setText(String.valueOf(snap));
    }

    // Handles the event of a user releasing the mouse button.
    public void mouseReleased(MouseEvent e){
	super.mouseReleased(e);
	positionLines();
    }
    
    /** Access function for the arc weight stored in the JDOM Document - default 1.
     * @return The Arc weight as stored in the Document - defaults to 1 if not present. 
     */
    public int getWeight(){
	// if <inscription><value> tag is empty or not present, assume
	// default arc weight of 1.
	int weight = 1;
	
	if (element.getChildren("inscription",ns).size()>0) {
	    if (element.getChild("inscription",ns).getChildren("value",ns).size()>0) {
		return Integer.parseInt(element.getChild("inscription",ns).getChildText("value",ns));
	    }
	}
	return weight;
    }

    /** Sets the Arc weight in the JDOM Document.
     */
    public void setWeight(int weight){
	if (element.getChildren("inscription",ns).size()<=0) {
	    element.addContent(new Element("inscription",ns));
	}
	if (element.getChild("inscription",ns).getChildren("value").size()<=0) {
	    element.getChild("inscription",ns).addContent(new Element("value",ns));
	}
	element.getChild("inscription",ns).getChild("value",ns).setText(String.valueOf(weight));
    }

    
    /** Returns the source ID stored in the Arc element */
    public String getSourceID(){
	return element.getAttributeValue("source");
    }
    
    /** Returns the target ID stored in the Arc element */
    public String getTargetID(){
	return element.getAttributeValue("target");
    }

    /** Sets the local reference to the Label representing the arc's Source. */
    public void setSourceLabel(NodeLabel source){
	sourcelabel = source;
	//System.out.println("ArcLabel:: set source label "+source.getName());
    }
    /** Sets the local reference to the Label representing the arc's Target. */
    public void setTargetLabel(NodeLabel target){
	targetlabel = target;
	//System.out.println("ArcLabel:: set target label "+target.getName());
    }
    
    /** Arc is active if its Source is active - it informs the source of the arc weight, and asks it whether it is active. */
    public boolean isActive(){
	return sourcelabel.isActive(getWeight());
    }
    
    /** When fired, an Arc adds its weight in Tokens to its target.
     * If the target is not a Place, an error message is printed. 
     * @see #unfire()
     */
    public void fire(){
	try {
	    PlaceLabel p = (PlaceLabel) targetlabel;
	    p.addToken(getWeight());
	} catch (Exception e){
	    System.out.println("ArcLabel::Can't add tokens to a non-place");
	}
    }
    
    /** When un-fired, an Arc removes its weight in Tokens from its source.
     * If the source is not a Place, an error message is printed.
     * @see #fire()
     */
    public void unfire(){
	try {
	    PlaceLabel p = (PlaceLabel) sourcelabel;
	    p.remToken(getWeight());
	} catch (Exception e){
	    System.out.println("ArcLabel::Can't add tokens to a non-place");
	}
    }

    // Function to create a new arc element.
    public static Element makeElement(Integer x, Integer y, String id, String source, String target, Namespace ns) {
	
	Element arc = new Element("arc", ns);
	arc.setAttribute("id", id);
	arc.setAttribute("source", source);
	arc.setAttribute("target", target);
	
	Element graphics = PNLabel.makeGraphics(String.valueOf(x), String.valueOf(y), ns);
	Element inscription = makeInscription(ns);
	
	arc.addContent(inscription);
	arc.addContent(graphics);
	
	return arc;

    }
    
    // function to create an inscription element used by makeElement
    public static Element makeInscription(Namespace ns) {
	
	Element inscription = new Element("inscription", ns);
	Element graphics = new Element("graphics", ns);
	Element offset = new Element("offset", ns);
	Element value = new Element("value", ns);
	
	offset.setAttribute("page", "1");
	offset.setAttribute("x", "20");
	offset.setAttribute("y", "0");
	value.setText("1");
	
	graphics.addContent(offset);
	inscription.addContent(graphics);
	inscription.addContent(value);
	
	return inscription;
    }
    
    /** Returns a list of PNLabels which are linked to this PNLabel. 
     * The list will consist of the Source label and the Target label. 
     */
    public java.util.List getLinkedLabels(){
	java.util.List linked = new java.util.LinkedList();
	linked.add(sourcelabel);
	linked.add(targetlabel);
	return linked;
    }
    
    /** A cleanup function that is called when the label is removed from the display.
     * It removes its LineLabel, and updates the reference to itself
     * stored within its source and target labels.
     */
    public DisplayPane remove(){
	DisplayPane display = super.remove();
	System.out.println("ArcLabel:: removing");
	
	display.remove(line);
	line = null;

	// update source/target labels of change.
	sourcelabel.removeOutputArc(this);
	targetlabel.removeInputArc(this);

	return display;
    }
    
    /** This overridden version also deals with highlighting the lines.
     */
    public void highlight(boolean highlight){
	super.highlight(highlight);
	highlightLine(highlight);
    }

    /** Highlights (or un-highlights) the line. */
    public void highlightLine(boolean highlight){
	if (line!=null)
	    line.highlight(highlight);
    }

    /** Re-use this function to change the arc weight. */
    public void addToken(){
	int weight = getWeight();
	setWeight(++weight);
	repaint();
    }

    /** Re-use this function to change the arc weight. */
    public void remToken(){
	int weight = getWeight();
	if (weight > 1){
	    setWeight(--weight);
	    repaint();
	}
    }
    
}



