package uk.ac.ic.doc.y2002.tool.editor;

import java.io.*;
import java.lang.reflect.*;
import javax.swing.JLabel;
import java.util.*;
import org.jdom.*;
import org.jdom.input.*;
import uk.ac.ic.doc.y2002.tool.application.*;

/** The ClassManager provides extensibility to the application,
 * by selecting the classes that should be used to represent a 
 * particular type of net. One ClassManager is loaded for each
 * display pane opened, and each determines the necessary classes
 * on its creation. It is given the type of the net being
 * displayed, and loads in the file "nettypes.xml" which defines the
 * net type hierarchy and the corresponding classes. This XML file may
 * be modified by the user to include support for new net types
 * and/or new *Label classes.
 * Another level of extensibility is managed by the Label classes
 * themselves - upon initialisation, each Label may specify a subclass
 * of the Right-click menu (context menu) and/or the PropertiesBox
 * which should be used.
 * @see DisplayPane#addLabels()
 */
public class ClassManager {  
    Class placelabel, transitionlabel, arclabel, 
	subnetlabel, referenceplacelabel, importplacelabel, exportplacelabel, 
	animator;
    static String defaulttype = "ptNet";

    /**
     * @param nettype the net type of the petri net being displayed. 
     *        eg in <net type="ptNet">, "ptNet" should be passed.
     */
    public ClassManager(String nettype) {
	if (nettype==null || nettype.length()==0) {
	    // net type is not specified in the XML file
	    System.out.println("ClassManager::type not specified! will use the default");
	    // set type to default
	    nettype = defaulttype;
	} 
	//else System.out.println("ClassManager::type of net is: "+nettype);
	loadHierarchy(nettype);
    }
    
    /** function which loads in the XML file specifying the nettype hierarchy,
     * and discovers a path through the hierarchy to the given nettype.
     * When the nettype is found, upon the return path later function pick
     * up the classes.
     * @param nettype the net type (eg "ptNet") to search for.
     * @see #findType(String thetype, Element root)
     */
    public void loadHierarchy(String nettype){
	//System.out.println("ClassManager::Searching for type hierarchy:");     
	// Request document building without validation
	SAXBuilder builder = new SAXBuilder(false);
	try {
	    Document doc = builder.build(new File("nettypes.xml"));  
	    // find the given nettype
	    Element root = doc.getRootElement(); // <pnmlnettypes>
	    if (!findType(nettype,root)) {
		System.out.println("Unknown net type: will be treated as the default type (see nettypes.xml)");
		findType(defaulttype,root);
	    }
	} catch (Exception e) {
	    System.out.println("ClassManager::Error loading in type hierarchy: "+e);
	}
	System.out.println();
    }

    /** Recursive function which searches the tree for the given type.
     *  When it finds the type, it gets the classes, from the leaf first
     *  and picks up other classes as it returns to the root.
     *  @see #getClasses(Element e) 
     *  @param thetype the type to search for
     *  @param root the Element at the root of the tree: <pnmlnettypes>
     */
    protected boolean findType(String thetype, Element root){
	java.util.List nettypes = root.getChildren("nettype");
	Iterator ts= nettypes.iterator();
	while (ts.hasNext()) {
	    Element t = (Element) ts.next();
	    String type = t.getAttributeValue("id");
	    //System.out.println("  -Found nettype: "+ type);
	    if (type.equals(thetype)){
		System.out.print("ClassManager:: Found the type hierarchy: "+type);
		getClasses(t);
		return true;
	    }
	    if (findType(thetype,t)) {
		System.out.print(" <- "+type);
		getClasses(t);
		return true;
	    }
	}
	return false;
    }
    
    /** This function extracts classes from a <nettype> element.
     *  It stores each class if one for that type hasn't yet been found.
     * @see #getClass(Element e, String elementname, Class eclass)
     * @param e the <nettype> element e to search for classes
     */
    protected void getClasses(Element e){
	//System.out.println(" ---Getting classes..." );
	placelabel = getClass(e,"placelabel",placelabel);
	transitionlabel = getClass(e,"transitionlabel",transitionlabel); 
	arclabel = getClass(e,"arclabel",arclabel); 
	subnetlabel = getClass(e,"subnetlabel",subnetlabel); 
	referenceplacelabel = getClass(e,"referenceplacelabel",referenceplacelabel); 
	importplacelabel = getClass(e,"importplacelabel",importplacelabel); 
	exportplacelabel = getClass(e,"exportplacelabel",exportplacelabel); 
	animator = getClass(e,"animator",animator); 
    }
    
    /** Extracts and returns a Class object from the hierarchy, given parameters.
     * @param e the Element <nettype> which may contain class names
     * @param elementname the sub-element's name, eg <placelabel>
     * @param eclass the Class object in ClassManager which keeps track of this type of class (eg a Place class)
     * @return a Class object for the class requested, or null if a) this particular type of class has already been specified, or b) class not defined (element 'elementname' not found)
     */
    protected Class getClass(Element e, String elementname, Class eclass){
	String fclass = e.getChildTextTrim(elementname);    
	if (fclass!=null && eclass==null) {
	    try {
		// note this doesn't actually change what eclass points to
		// *outside* this function - it's passed by value! 
		eclass = Class.forName(fclass);
		//System.out.println("  --Found class: "+eclass.getName());
	    } catch (Exception ex) {
		System.out.println("ClassManager:: Error loading in class "+elementname+" : "+ex);
		return null;
	    }
	}
	return eclass;
    }
    
    /** Used by the more specific new*Label functions.
     */
    protected Object newPNLabel(Element e, Class c){
	Object o = null; 
	try {
	    o = c.getConstructor(new Class[] {Element.class})
		.newInstance(new Object[] {e});
	    //    System.out.println("ClassManager::Successfully created new PNLabel");
	} catch (Exception ex) {
	    System.out.println("ClassManager::Error creating PNLabel! "+c.getName());
	}
	return o;
    }
    /** Call this function whenever a new PlaceLabel needs to be created.
     * It returns an object nominally of class PlaceLabel, but it may actually be
     * a more specific subclass of PlaceLabel, determined by the ClassManager.
     */
    public PlaceLabel newPlaceLabel(Element e){
	return (PlaceLabel) newPNLabel(e,placelabel);
    }

    /** @see #newPlaceLabel(Element e) */
    public TransitionLabel newTransitionLabel(Element e){
	return (TransitionLabel) newPNLabel(e,transitionlabel);
    }
    
    /** @see #newPlaceLabel(Element e) */
    public ArcLabel newArcLabel(Element e){
	return (ArcLabel) newPNLabel(e,arclabel);
    }
    
    /** @see #newPlaceLabel(Element e) */
    public SubnetLabel newSubnetLabel(Element e){
	return (SubnetLabel) newPNLabel(e,subnetlabel);
    }
    
    /** @see #newPlaceLabel(Element e) */
    public ReferencePlaceLabel newReferencePlaceLabel(Element e){
	return (ReferencePlaceLabel) newPNLabel(e,referenceplacelabel);
    }
    /** @see #newPlaceLabel(Element e) */
    public ImportPlaceLabel newImportPlaceLabel(Element e){
	return (ImportPlaceLabel) newPNLabel(e,importplacelabel);
    }
    /** @see #newPlaceLabel(Element e) */
    public ExportPlaceLabel newExportPlaceLabel(Element e){
	return (ExportPlaceLabel) newPNLabel(e,exportplacelabel);
    }

    public Element newPlaceElement(Object[] arguments) {
	return newElement(arguments, placelabel);
    }

    public Element newReferencePlaceElement(Object[] arguments) {
	return newElement(arguments, referenceplacelabel);
    }

    public Element newTransitionElement(Object[] arguments) {
	return newElement(arguments, transitionlabel);
    }
    
    public Element newSubnetElement(Object[] arguments) {
	return newElement(arguments, subnetlabel);
    }
    
    public Element newArcElement(Object[] arguments) {
	return newElement(arguments, arclabel);
    }
    
    // TODO  new subnet element and new refereneceplaceelement
         
    public Element newElement(Object[] arguments, Class c) {
	
	int length = 0;
	Class[] parameterTypes;
	Element newElement = null;
	List object_list = Arrays.asList(arguments);

	// get length
	Iterator it = object_list.iterator();
	while (it.hasNext()) {
	    it.next();
	    length++;
	}
   
	// initialize Class[]
	parameterTypes = new Class[length];
	for (int i=0; i <= length - 1; i++)
	    parameterTypes[i] = arguments[i].getClass();

	try {
	    Method makePlace = c.getMethod("makeElement", parameterTypes);
	    newElement = (Element) makePlace.invoke(null, arguments);
	        
	} catch (NoSuchMethodException e) {
	    System.out.println (e);
	}  catch (IllegalAccessException e) {
	    System.out.println(e);
	} catch (InvocationTargetException e) {
	    System.out.println(e);
	}
	
	return newElement;
    }
 
    /** @see #newPlaceLabel(Element e) */
    public PNAnimator newAnimator(DisplayPane display){
	Object o = null; 
	try {
	    o = animator.getConstructor(new Class[] {DisplayPane.class})
		.newInstance(new Object[] {display});
	    //    System.out.println("ClassManager::Successfully created new PNAnimator");
	} catch (Exception ex) {
	    System.out.println("ClassManager::Error creating PNAnimator!");
	}

	return (PNAnimator) o;
    }
}










