package uk.ac.ic.doc.y2002.tool.editor;

import javax.swing.*;
import java.lang.reflect.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.awt.image.*; 
import java.awt.*;
import java.util.*;
import org.jdom.*;
import uk.ac.ic.doc.y2002.tool.application.*;

/**
 * The <b>PNLabel</b> class is the basis of all Petri Net Labels in the program.
 * Each PNLabel contains a reference to the Element within the 
 * DOM that the Label represents on screen.
 * PNLabel itself should not be instantiated.
 * @see PlaceLabel
 */

public class PNLabel extends JLabel implements MouseListener, MouseMotionListener{
    /** Reference to the JDOM Element which represents this object. 
     *  All useful information about this object should be stored in
     *  and accessible from this Element.
     */
    protected Element element;
    /** Namespace of the JDOM Element which represents this object. */
    protected Namespace ns;
    /** Class that is used to generate a properties box for this 
     *  type of PNLabel. */
    protected Class propertiesbox;
    /** Class that is used to generate a context (right-click) menu for this 
     *  type of PNLabel. */
    protected Class contextmenu;
    Dimension preferredSize = new Dimension(40,40);
    protected boolean beingMoved=false;
    protected boolean selected=false;
    protected boolean firsttime=true;
    protected Color fillcolour = Color.white;
    protected Color normalfill = Color.white;
    protected Color highlightfill = new Color((float) 0.7,(float) 0.7,(float) 1);
    protected int height, width, lastX, lastY,
	offset = 0, xoffset = 0, yoffset = 0,
	scaling = 1, textspace = 25;
    protected int xPress, yPress;
    /** Locally-stored copy of the object's unique id. */
    protected String id;

    protected int rotation = 0; // rotation in degrees
    // currently in PNLabel because the right-click menu has Rotate
    // option for everything; move it to TransitionLabel later?

    /** @param element a JDOM Element which contains
     * all the relevant data for this Petri net object,
     * e.g. the &lt;place&gt; element.
     */
    public PNLabel(Element element) {
	super("");
	setHorizontalAlignment(CENTER);
	this.element = element;
	ns = element.getNamespace();
	id = element.getAttributeValue("id");
	contextmenu = PetriContextMenu.class;
	propertiesbox = PNPropertiesBox.class;
	addMouseMotionListener(this);
	addMouseListener(this);
    }
    
    /** The function which draws the PNLabel on screen.
     * Overridden in subclasses of PNLabel.
     */
    public void paintComponent(Graphics g){
        super.paintComponent(g);
	if (firsttime) {
	    dofirst();
	}
    }

    /** An access function that returns the JDOM Element which this PNLabel represents.
     * @return the JDOM Element object for this PNLabel. 
     */
    public Element returnElement() {
	return element;
    }

    /** Changes the element this PNLabel is representing.
     * Not currently used in the program.
     */
    public void changeElement(Element newElement) {
	element = newElement;
    }

    /** @return true if this PNLabel is selected, false if not. */
    public boolean returnSelected() {
	return selected;
    }

    /** Should not be called directly: call select instead.
     * Deals with highlighting: changes the displayed colour of
     * the PNLabel, and sets the boolean selected. 
     */
    public void highlight(boolean highlight){
	if (highlight)
	    fillcolour = highlightfill;
	else fillcolour = normalfill;
	selected = highlight;
    }

    /** Selects the PNLabel.
     * Uses the PNAnimator associated with the current display
     * to deal with highlighting. 
     * @see PNAnimator
     */
    public void select(boolean select){
	//System.out.println("PNLabel::selecting");
	selected = select;
        getDisplayPane().animator.highlightPNLabel(this, select);
    }

    /** Selects the PNLabel without unselecting others in the display.
     * @see PNAnimator
     */
    public void selectOnly(boolean select) {
	selected = select;
	getDisplayPane().animator.highlightPNLabelOnly(this, select);
    }

    public Dimension getPreferredSize() {
        return preferredSize;
    }
    
    protected void dofirst() {
	updatePosition();
	firsttime=false;
    }

    // retrieves the X,Y position from the JDOM Document
    // and stores in lastX, lastY
    protected void updatePosition() {
	lastX = getJX();
	lastY = getJY();
	updateBounds();
    }

    // Updates the coordinates representing the location of the current label
    // in response to a Mouse event
    protected void updateLocation(MouseEvent e){
	setPosition(getX() + e.getX() - xPress, getY() + e.getY() - yPress);
	repaint();
    }
    protected void updateLocation(int newx, int newy){
	setPosition(newx, newy);
	repaint();
    }
    
    /** Update DOM with new values and set the label position.
     * The parameters given are the x,y coordinates for the screen position,
     * which may vary from the coordinates stored in the JDOM: this is because 
     * the XML file stores the centre position of the element, while the 
     * JLabel refers to the top left corner.
     * @param x the x position on screen
     * @param y the y position on screen
     */
    public void setPosition(int x, int y) {
	setLocation(x,y);
	setJX(x+xoffset);
	setJY(y+yoffset);
    }

    /** Get X value from the JDOM Document 
     * Defaults to 0 if not found
     */
    public int getJX() {
	String sint="0";
	if (element.getChildren("graphics",ns).size()>0)
	    if (element.getChild("graphics",ns).getChildren("position",ns).size()>0)
		sint = element.getChild("graphics",ns).getChild("position",ns).getAttributeValue("x");
	if (sint==null) sint="0";
	if (sint.equals("")) sint="0";
	return Integer.parseInt(sint);
    }
    /** Get Y value from the JDOM Document
     * Defaults to 0 if not found
     */
    public int getJY() {
	String sint="0";
	if (element.getChildren("graphics",ns).size()>0)
	    if (element.getChild("graphics",ns).getChildren("position",ns).size()>0)
		sint = element.getChild("graphics",ns).getChild("position",ns).getAttributeValue("y");
	if (sint==null) sint="0";
	if (sint.equals("")) sint="0";
	return Integer.parseInt(sint);
    }

    /** Set X value in the JDOM Document
     * note this may need conversion from the actual x,y values for the label
     */
    protected void setJX(int x) {
	if (element.getChildren("graphics",ns).size()<=0)
	    element.addContent(new Element("graphics",ns));
	if (element.getChild("graphics",ns).getChildren("position",ns).size()<=0)
	    element.addContent(new Element("position",ns));
	
	element.getChild("graphics",ns).getChild("position",ns).setAttribute("x", String.valueOf(x));
    }
    /** Set Y value in the JDOM Document
     * note this may need conversion from the actual x,y values for the label
     */
    protected void setJY(int y) {
	if (element.getChildren("graphics",ns).size()<=0)
	    element.addContent(new Element("graphics",ns));
	if (element.getChild("graphics",ns).getChildren("position",ns).size()<=0)
	    element.addContent(new Element("position",ns));
	
	element.getChild("graphics",ns).getChild("position",ns).setAttribute("y", String.valueOf(y));
    }

    /** may be overridden to deal with different sizes and
     * shapes of labels.
     */
    protected void updateBounds() {
	setBounds(lastX,lastY,40,40);
    }
    
    // Handles the event of a user dragging the mouse while holding
    // down the mouse button.
    public void mouseDragged(MouseEvent e){
	if ( beingMoved ) {
	    updateLocation(e); 
	    if (!selected)
		select(true);
	}
    }
    
    // Handles the event of the user pressing down the mouse button.
    public void mousePressed(MouseEvent e){
	if (e.isPopupTrigger()) {
	    //    System.out.println("Mouse pressed - popup trigger");
	    select(true);
	    showContextMenu(e);
	} else if (e.getID()==MouseEvent.MOUSE_CLICKED) {
	    //System.out.println("Mouse pressed - Mouse Clicked");
	    mouseClicked(e);
	} else if (e.getID()==MouseEvent.MOUSE_PRESSED) {
	    // System.out.println("Mouse pressed - Mouse Pressed on "+getID()+" "+e.getClickCount()+" times");
	    // if pointer selected in application.
	    if (Application.setTool.equals("pointer")) {
		beingMoved=true;
		lastX = getX() - e.getX();
		lastY = getY() - e.getY();
		xPress = e.getX();
		yPress = e.getY();
		updateLocation(e);
	    } else if (Application.setTool.equals("addtok")){
		addToken();
	    } else if (Application.setTool.equals("remtok")){
		remToken();
	    }
	    // else notify DisplayPane of source/parent
	    else {
		DisplayPane my_parent = getDisplayPane();
		my_parent.notify(this);
		
	    }
	}
    }
    
    // Handles the event of a user releasing the mouse button.
    public void mouseReleased(MouseEvent e){
	if ( beingMoved ) {
	    updateLocation(e);
	    getParent().repaint(); //update the display
	    beingMoved=false;
	} else {
	    // not being moved
	}
    }
    
    // Detects different types of clicks
    // Add actions as required
    public void mouseClicked(MouseEvent e){
	if (e.getClickCount()==1) {
	    // single clicked (with any button)
	    if ( (e.getModifiers() & InputEvent.BUTTON1_MASK) == InputEvent.BUTTON1_MASK) { 
		// Select/unselect label
		//System.out.println("Left button clicked");
		if (Application.setTool.equals("pointer")) {
		    select(!selected);
		}
	    } else if ( (e.getModifiers() & InputEvent.BUTTON2_MASK) == InputEvent.BUTTON2_MASK) {
		// middle button clicked
		// ...
	    } else if ( (e.getModifiers() & InputEvent.BUTTON3_MASK) == InputEvent.BUTTON3_MASK) {
		// right button clicked
		// opens right-click context menu
		//System.out.println("Right button clicked");
		select(true);
		showContextMenu(e);
	    }
	} else if (e.getClickCount()==2) {
	    // double clicked (with any button)
	    doubleclickAction();
	}
    }
    
    /** Override in subclasses to produce required functionality */
    public void doubleclickAction(){
    }
    
    /** Brings up the right-click popup menu.
     *  The class is determined by the local variable propertiesbox. */
    public void showContextMenu(MouseEvent e) {
	try {
	    PetriContextMenu popup; 
	    Class[] argtypes = new Class[] {PNLabel.class,Class.class};
	    Constructor constructor = contextmenu.getConstructor(argtypes);
	    Object[] args = new Object[] {this, propertiesbox};
	    popup = (PetriContextMenu) constructor.newInstance(args);
	    popup.show(this, e.getX(), e.getY());
	    //System.out.println("PNLabel::Successfully created context menu");
	} catch (Exception ex) {
	    System.out.println("PNLabel::Error creating context menu!");
	}
    }
    
    // This method is required by MouseListener.
    public void mouseMoved(MouseEvent e){}
    
    // These methods are required by MouseMotionListener.
    public void mouseExited(MouseEvent e){}
    public void mouseEntered(MouseEvent e){}
    

    /** @return The DisplayPane which contains this PNLabel. */
    public DisplayPane getDisplayPane(){
	// select the parent JDesktopFrame and add the Properties popup to it
	Container con = getParent();
	Class c = con.getClass();
	if (c == DisplayPane.class) {
	    //    System.out.println("PNLabel::got parent display pane");
	    return (DisplayPane) con;
	} else {
	    // shouldn't happen!
	    System.out.println("PNLabel::Error - can't find parent display pane!");
	    return null;
	}
    }
    
    /** Creates a graphics JDOM Element for use by a new PN element. 
     * @return &lt;graphics&gt; element corresponding to this PNLabel's current values 
     */
    public static Element makeGraphics(String x, String y, Namespace ns) {
	
	Element graphics = new Element("graphics", ns);
	Element g_position = new Element("position", ns);
	g_position.setAttribute("page", "1");
	g_position.setAttribute("x", x);
        g_position.setAttribute("y", y);
	graphics.addContent(g_position);
	
	return graphics;
    }
    
    /** @return locally-stored unique id of the Petri net object */
    public String getID() {
	return id;
    }

    /* @return name of the basic element type, e.g. &lt;place&gt; will return "place" */
    public String myElementName() {
	return element.getName();
    }
    
    /** Function that should be called when the element is removed from the Petri Net. 
     *  May override in subclasses to do custom display cleanup as necessary.
     *  Removes both the element from the JDOM and the PNLabel from the DisplayPane that contains it.
     *  @return The DisplayPane that contained the PNLabel.
     */
    public DisplayPane remove(){

	// remove element from JDOM.
	Element my_parent = element.getParent();
	my_parent.removeContent(element);

	// remove PNLabel from DisplayPane it is in
	DisplayPane display = getDisplayPane();
	display.remove(this);

	return display;
    }
    
    /** Override in subclasses to indicate when an element is enabled within the net. */
    public boolean isActive(){
        return false;
    } 

    /** Override this as necessary (eg for PlaceLabel). 
     * @see PlaceLabel
     * @see ArcLabel
     */
    public void addToken(){
    }
    
    /** Override as necessary (eg for PlaceLabel).
     * @see PlaceLabel
     * @see ArcLabel
     */
    public void remToken(){
    }

    /** Override in subclasses to return a list of PNLabels
     * which are linked to this PNLabel.
     * @return a List of PNLabels which are linked to this PNLabel. */
    public java.util.List getLinkedLabels(){
	java.util.List linked = new java.util.LinkedList();
	return linked;
    }

    /** Function which allows the label to update itself (eg
     * for modifying screen display) in response to changes
     * in the underlying JDOM Element. Override in subclasses.
     */
    public void updateSelf(){
    }

    /** Function to find and return the Toolspecific element 
     * corresponding to this tool. If not found, returns null.
     */
    public Element getToolspecific(){
	Element toolspecific=null;
	java.util.List toolspecifics = element.getChildren("toolspecific",ns);
	if (toolspecifics.size()>0) {
	    Iterator ts = toolspecifics.iterator();
	    while (ts.hasNext()) {
		Element t = (Element) ts.next();
		if (t.getAttributeValue("tool").equals("EXPT")){
		    toolspecific = t;
		    break;
		}
	    }
	}
	/*
	    if (toolspecific!=null) System.out.println("PNLabel:: found toolspecific tag for element "+getID());
	      else System.out.println("PNLabel:: not found toolspecific tag for element "+getID());
	*/
	return toolspecific;
    }

    /** Function to make a Toolspecific element - assumes
     * one does not exist already. (check with getToolspecific first)
     * Returns the new <toolspecific> element.
     */
    public Element makeToolspecific(){
	Element toolspecific = new Element("toolspecific",ns);
	element.addContent(toolspecific);
	toolspecific.setAttribute("tool","EXPT");
	toolspecific.setAttribute("version","1.0");
	return toolspecific;
    }
}




